use cgmath::Deg;
use std::time::Instant;
use winit::event::{ElementState, MouseButton, MouseScrollDelta, TouchPhase, VirtualKeyCode};

use super::{Camera, InputStatus, Ship, System};
use crate::{consts, content::Content, render::Sprite, render::Spriteable};

pub struct Game {
	pub input: InputStatus,
	pub last_update: Instant,
	pub player: Ship,
	pub test: Ship,
	pub system: System,
	pub camera: Camera,
	paused: bool,
	pub time_scale: f32,
}

impl Game {
	pub fn new(ct: Content) -> Self {
		Game {
			last_update: Instant::now(),
			input: InputStatus::new(),
			player: Ship::new(&ct.ships[0], (0.0, 0.0).into()),
			camera: Camera {
				pos: (0.0, 0.0).into(),
				zoom: 500.0,
			},
			system: System::new(&ct.systems[0]),
			paused: false,
			time_scale: 1.0,
			test: Ship::new(&ct.ships[0], (100.0, 100.0).into()),
		}
	}

	pub fn process_key(&mut self, state: &ElementState, key: &VirtualKeyCode) {
		self.input.process_key(state, key)
	}

	pub fn process_click(&mut self, state: &ElementState, key: &MouseButton) {
		self.input.process_click(state, key)
	}

	pub fn process_scroll(&mut self, delta: &MouseScrollDelta, phase: &TouchPhase) {
		self.input.process_scroll(delta, phase)
	}

	pub fn set_paused(&mut self, pause: bool) {
		if pause {
			self.paused = true;
			self.input.release_all()
		} else {
			self.paused = false;
		}
	}

	pub fn update(&mut self) {
		let t: f32 = self.last_update.elapsed().as_secs_f32() * self.time_scale;

		self.player.engines_on = self.input.key_thrust;
		if self.input.key_thrust {
			self.player.physicsbody.thrust(50.0 * t);
		}

		if self.input.key_right {
			self.player.physicsbody.rot(Deg(35.0) * t);
		}

		if self.input.key_left {
			self.player.physicsbody.rot(Deg(-35.0) * t);
		}

		if self.input.v_scroll != 0.0 {
			self.camera.zoom =
				(self.camera.zoom + self.input.v_scroll).clamp(consts::ZOOM_MIN, consts::ZOOM_MAX);
			self.input.v_scroll = 0.0;
		}

		self.player.physicsbody.tick(t);
		self.camera.pos = self.player.physicsbody.pos;

		self.last_update = Instant::now();
	}

	pub fn get_sprites(&self) -> Vec<Sprite> {
		let mut sprites: Vec<Sprite> = Vec::new();

		sprites.append(&mut self.system.get_sprites());
		sprites.push(self.player.get_sprite());
		sprites.push(self.test.get_sprite());

		// Make sure sprites are drawn in the correct order
		// (note the reversed a, b in the comparator)
		//
		// TODO: maybe use a gpu depth buffer instead?
		// I've tried this, but it doesn't seem to work with transparent textures.
		sprites.sort_by(|a, b| b.pos.z.total_cmp(&a.pos.z));

		return sprites;
	}
}
